/*
 * an experimental digital detector
 *
 * by T.INOUE (JI3GAB)
 *
 *  Copyright (C) 2000-2006 Takafumi INOUE
 *
 *
 * center frequency: 12kHz (default)
 * sample rate: 48kHz
 *
 * 2000.4.25
 * Java version
 *
 * Java version Last modified 2000.10.13
 *
 * 2001.01.30
 * C++ again version
 *
 * 2005.11
 * now can change filter bw on demand
 * 
 * 2006.2
 * add BFO 
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "Detector.h"

#define BUFFER_SIZE 256
#define DECIMATION_FACTOR 4

void
Detector::detect(float *input, float *output, int n_samples) {
  int decimation_factor = DECIMATION_FACTOR;

  // I/Q buffer for higher sampling rate (48kHz)
  static float ii[BUFFER_SIZE];
  static float qq[BUFFER_SIZE];
  // I/Q buffer for lower sampling rate (12kHz)
  static float iiLow[BUFFER_SIZE];
  static float qqLow[BUFFER_SIZE];
  // buffer for demodulated output at lower sampling rate
  static float output_low[BUFFER_SIZE];

  // bfo buffer
  static float ibfo[BUFFER_SIZE], qbfo[BUFFER_SIZE];
  
  //in_Filter->do_filter(input, n_samples);

  bfo->get_sample_n(ibfo, qbfo, n_samples);
  
  int n_ii = 0;
  int n_qq = 0;

  // mixer (down conversion to baseband)
  for (int i = 0; i < n_samples; i++) {
    ii[n_ii++] = input[i] * ibfo[i];
    qq[n_qq++] = input[i] * qbfo[i];
  }

  // decimation (to 12kHz sampling rate)
  iDecimator->decimate(ii, n_samples, iiLow);
  qDecimator->decimate(qq, n_samples, qqLow);
  int n_samples_low = n_samples / decimation_factor;

  if (mode == 1 || mode ==2) {
    // I/Q filter
    // I signal is passed through a BPF
    iFilter->do_filter(iiLow, n_samples_low);
    // Q signal is passed through a BPF which has the
    // same amplitude response as the one for I signal
    // but also has 90 degree phase shifter built in.
    qFilter->do_filter(qqLow, n_samples_low);
  }

  for (int i = 0; i < n_samples_low; i++) {
    float out;
    if (mode == 0) {
      // AM envelope detection
      out = (float)(sqrt(iiLow[i] * iiLow[i] + qqLow[i] * qqLow[i]) * 1);
      // out = m[i];
    } else if (mode == 1) {
      // LSB
      out = iiLow[i] + qqLow[i];
    } else {
      // USB
      out = iiLow[i] - qqLow[i];

    }
    output_low[i] = out;
  }

  // interpolation (to 48kHz sampling rate)
  interpolator->interpolate(output_low, n_samples_low, output);
}
    
void
Detector::set_iqfilter(float sf, float f1, float f2, float trband, float attn) {
  if (iFilter != NULL) {
    delete iFilter;
  }
  if (qFilter != NULL) {
    delete qFilter;
  }
  iFilter = new FIRFilter(IFIL, sf, f1, f2, trband, attn);
  //iFilter->filterGain();
  qFilter = new FIRFilter(QFIL, sf, f1, f2, trband, attn);
  //qFilter->filterGain();
}

void Detector::filter_init() {
      
  // set up filters
  float *a = (float *)(malloc(sizeof(float) * 89));
  float *b = (float *)(malloc(sizeof(float) * 45));
      
  /* Kaiser Window FIR Filter
   * Sampling Rate: 48000Hz
   * Filter type: Low pass
   * Passband: 0.0 - 4000.0 Hz
   * Order: 88
   * Transition band: 2000.0 Hz
   * Stopband attenuation: 60.0 dB
   * 
   * Coefficients:
   */
  a[0] =	-7.36821E-5f;
  a[1] =	2.7408667E-5f;
  a[2] =	2.0153754E-4f;
  a[3] =	3.7063626E-4f;
  a[4] =	4.1388065E-4f;
  a[5] =	2.2914718E-4f;
  a[6] =	-1.9103645E-4f;
  a[7] =	-7.121183E-4f;
  a[8] =	-0.0010790983f;
  a[9] =	-0.0010190749f;
  a[10] =	-3.923451E-4f;
  a[11] =	6.794222E-4f;
  a[12] =	0.0017886907f;
  a[13] =	0.0023678844f;
  a[14] =	0.0019423079f;
  a[15] =	4.103589E-4f;
  a[16] =	-0.0017912489f;
  a[17] =	-0.0037567662f;
  a[18] =	-0.004442365f;
  a[19] =	-0.0031565337f;
  a[20] =	-3.9379997E-9f;
  a[21] =	0.003979839f;
  a[22] =	0.007066775f;
  a[23] =	0.0075503704f;
  a[24] =	0.0045579057f;
  a[25] =	-0.0013257348f;
  a[26] =	-0.007996592f;
  a[27] =	-0.012480278f;
  a[28] =	-0.012136741f;
  a[29] =	-0.0059750136f;
  a[30] =	0.004508162f;
  a[31] =	0.015448019f;
  a[32] =	0.021831708f;
  a[33] =	0.019497888f;
  a[34] =	0.00719949f;
  a[35] =	-0.012135523f;
  a[36] =	-0.031611286f;
  a[37] =	-0.042199712f;
  a[38] =	-0.035728954f;
  a[39] =	-0.008031254f;
  a[40] =	0.038919125f;
  a[41] =	0.09677941f;
  a[42] =	0.15278521f;
  a[43] =	0.1933482f;
  a[44] =	0.20815012f;
  a[45] =	0.1933482f;
  a[46] =	0.15278521f;
  a[47] =	0.09677941f;
  a[48] =	0.038919125f;
  a[49] =	-0.008031254f;
  a[50] =	-0.035728954f;
  a[51] =	-0.042199712f;
  a[52] =	-0.031611286f;
  a[53] =	-0.012135523f;
  a[54] =	0.00719949f;
  a[55] =	0.019497888f;
  a[56] =	0.021831708f;
  a[57] =	0.015448019f;
  a[58] =	0.004508162f;
  a[59] =	-0.0059750136f;
  a[60] =	-0.012136741f;
  a[61] =	-0.012480278f;
  a[62] =	-0.007996592f;
  a[63] =	-0.0013257348f;
  a[64] =	0.0045579057f;
  a[65] =	0.0075503704f;
  a[66] =	0.007066775f;
  a[67] =	0.003979839f;
  a[68] =	-3.9379997E-9f;
  a[69] =	-0.0031565337f;
  a[70] =	-0.004442365f;
  a[71] =	-0.0037567662f;
  a[72] =	-0.0017912489f;
  a[73] =	4.103589E-4f;
  a[74] =	0.0019423079f;
  a[75] =	0.0023678844f;
  a[76] =	0.0017886907f;
  a[77] =	6.794222E-4f;
  a[78] =	-3.923451E-4f;
  a[79] =	-0.0010190749f;
  a[80] =	-0.0010790983f;
  a[81] =	-7.121183E-4f;
  a[82] =	-1.9103645E-4f;
  a[83] =	2.2914718E-4f;
  a[84] =	4.1388065E-4f;
  a[85] =	3.7063626E-4f;
  a[86] =	2.0153754E-4f;
  a[87] =	2.7408667E-5f;
  a[88] =	-7.36821E-5f;

/*
 * Kaiser Window FIR Filter
 * Filter type: Low pass
 * Passband: 0.0 - 16000.0 Hz
 * Order: 44
 * Transition band: 4000.0 Hz
 * Stopband attenuation: 60.0 dB
 */

  b[0] =	2.9470463E-4f;
  b[1] =	-4.0304553E-4f;
  b[2] =	-1.1397117E-10f;
  b[3] =	0.0010437695f;
  b[4] =	-0.002158033f;
  b[5] =	0.0021436461f;
  b[6] =	3.940453E-10f;
  b[7] =	-0.003884317f;
  b[8] =	0.007164466f;
  b[9] =	-0.0065035834f;
  b[10] =	-8.341263E-10f;
  b[11] =	0.010345707f;
  b[12] =	-0.018230205f;
  b[13] =	0.015992f;
  b[14] =	1.3368491E-9f;
  b[15] =	-0.024631264f;
  b[16] =	0.043660108f;
  b[17] =	-0.039335676f;
  b[18] =	-1.7409759E-9f;
  b[19] =	0.07145248f;
  b[20] =	-0.15566476f;
  b[21] =	0.22367613f;
  b[22] =	0.7492836f;
  b[23] =	0.22367613f;
  b[24] =	-0.15566476f;
  b[25] =	0.07145248f;
  b[26] =	-1.7409759E-9f;
  b[27] =	-0.039335676f;
  b[28] =	0.043660108f;
  b[29] =	-0.024631264f;
  b[30] =	1.3368491E-9f;
  b[31] =	0.015992f;
  b[32] =	-0.018230205f;
  b[33] =	0.010345707f;
  b[34] =	-8.341263E-10f;
  b[35] =	-0.0065035834f;
  b[36] =	0.007164466f;
  b[37] =	-0.003884317f;
  b[38] =	3.940453E-10f;
  b[39] =	0.0021436461f;
  b[40] =	-0.002158033f;
  b[41] =	0.0010437695f;
  b[42] =	-1.1397117E-10f;
  b[43] =	-4.0304553E-4f;
  b[44] =	2.9470463E-4f;


  // set up decimator and interpolator
  iDecimator = new Decimator(a, 89, 4);
  qDecimator = new Decimator(a, 89, 4);

  interpolator = new Interpolator(a, 89, 4);

  // input filter
  in_Filter = new FIRFilter(b, 45);

  // analytic filter pair
  iFilter = new FIRFilter(IFIL, 12000.0f, 200.0f, 3000.0f, 200.0f, 60.0f);
  qFilter = new FIRFilter(QFIL, 12000.0f, 200.0f, 3000.0f, 200.0f, 60.0f);
}

void
Detector::set_ifreq(float freq) {
  bfo->set_freq(freq);
}
