/*
 * FIRFilter class
 *
 * by JI3GAB (ji3gab@arrl.net)
 *
 * The filter design function is modified version of the FIR filter design
 * applet source code
 * on the "Digital Signal Processing Tutorial" site (www.dsptutor.freeuk.com).
 * MNY TNX to the original author.
 *
 *
 */

#ifndef FIR_H
#define FIR_H

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#define LPF 0
#define HPF 1
#define BPF 2
#define IFIL 3
#define QFIL 4
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

//#define DEBUG_OUTPUT
#ifdef DEBUG_OUTPUT
#define DOUT(string, ...) \
 fprintf(stderr, "debug:" string, __VA_ARGS__)
#else
#define DOUT(string, ...) {}
#endif

class FIRFilter {
  
private:
  float *coeff;			// filter coefficients
  int n_tap;			// number of taps
  int in_idx;			// indicates where the next data is to be placed
  float *buf;			// used as a circular buffer
  
  float I0(float x) {
    // zero order Bessel function of the first kind
    float eps = 1.0e-6f;   // accuracy parameter
    float fact = 1.0f;
    float x2 = 0.5f * x;
    float p = x2;
    float t = p * p;
    float s = 1.0f + t;
    for (int k = 2; t > eps; ++k) {
      p *= x2;
      fact *= k;
      t = sqr(p / fact);
      s += t;
    }
    return s;
  }
  float sqr(float x) {
    return x * x;
  }
  
  void filterGain () {
    float *a = coeff;
    int freqPoints = 250;
    // filter gain at uniform frequency intervals
    float *g = (float *)malloc(sizeof(float) * (freqPoints + 1));
    float theta, s, c, sac, sas;
    float gMax = -100.0f;
    float sc = 10.0f / (float)log(10.0f);
    float t = M_PI / freqPoints;
    for (int i = 0; i <= freqPoints; i++) {
      theta = i * t;
      sac = 0.0f;
      sas = 0.0f;
      for (int k = 0; k < n_tap; ++k) {
	c = (float)cos(k * theta);
	s = (float)sin(k * theta);
	sac += c * a[k];
	sas += s * a[k];
      }
      g[i] = sc * (float)log(sqr(sac) + sqr(sas));
      gMax = fmax(gMax, g[i]);
    }
    // normalise to 0 dB maximum gain
    for (int i = 0; i <= freqPoints; ++i) {
	g[i] -= gMax;
    }
    // normalise coefficients
    float normFactor = (float)pow(10.0, -0.05 * gMax);
    for (int i = 0; i < n_tap; ++i) {
      a[i] *= normFactor;
    }
#ifdef DEBUG_OUTPUT
    for (int i = 0; i < freqPoints; ++i) {
      DOUT("g[%d]: = %.7e\n", i, g[i]);
    }
#endif
    free(g);
  }

public:
  FIRFilter(float *a, int n) {

    in_idx = 0;
    n_tap = n;
    coeff = a;
    buf = (float *)(malloc(sizeof(float) * n_tap));
  
  }
  FIRFilter(int filterType, float rate, float f1, float f2, float trband, float atten) {

    float fN = rate / 2;
    float alpha = 0.0f;
    float modfreq = 0.0f;
    
    // estimate filter order
    int order = 2 * (int) ((atten - 7.95) / (14.36 * trband / fN) + 1.0f);
    // estimate Kaiser window parameter (alpha):
    if (atten >= 50.0f) {
      alpha = 0.1102f * (atten - 8.7f);
    } else if (atten > 21.0f){
      alpha = 0.5842f * (float)exp(0.4f * (float)log(atten - 21.0f))
	+ 0.07886f * (atten - 21.0f);
    } else {
      alpha = 0.0f;
    }
    // window function values
    float I0alpha = I0(alpha);
    int m = order / 2;
    float *win = (float *)(malloc(sizeof(float) * (m + 1)));
    
    for (int n=1; n <= m; ++n) {
      win[n] = I0(alpha * (float)sqrt(1.0f - sqr((float)n/m))) / I0alpha;
    }
    float w0 = 0.0f;
    float w1 = 0.0f;

    switch (filterType) {
    case HPF: w0 =M_PI;
      w1 =M_PI * (1.0f - (f1 - 0.5f * trband) / fN);
      break;
    case BPF: w0 = 0.5f * M_PI * (f1 + f2) / fN;
      w1 = 0.5f * M_PI * (f2 - f1 + trband) / fN;
      break;
    case IFIL: f2 = (f2 - f1) / 2;
      modfreq = f1 + f2;
      w0 = 0.0f;
      w1 = M_PI * (f2 + 0.5f * trband) / fN;
      break;
    case QFIL: f2 = (f2 - f1) / 2;
      modfreq = f1 + f2;
      w0 = 0.0f;
      w1 = M_PI * (f2 + 0.5f * trband) / fN;
    default: w0 = 0.0f;
      w1 = M_PI * (f2 + 0.5f * trband) / fN;
    }
    DOUT("modfreq: %e\n", modfreq);
    // filter coefficients (NB not normalised to unit maximum gain)
    // desgin Low pass filter
    float *a = (float *)(malloc(sizeof(float) * (order + 1)));
    a[0] = w1 / M_PI;
    for (int n = 1; n <= m; ++n) {
      a[n] = (float)sin(n * w1) * (float)cos(n * w0) * win[n] / (n * M_PI);
    }
    free(win);
    // shift impulse response to make filter causal:
    for (int n = m + 1; n <= order; ++n) {
      a[n] = a[n - m];
    }
    for (int n = 0; n <= m-1; ++n) {
      a[n] = a[order - n];
    }
    a[m] = w1 / M_PI;
    
    // modulate coeffs to shift center freq and to
    // get analytic pair filter
    //   added  by T.INOUE
    for (int n = 0; n <= order; ++n) {
      if (filterType == IFIL) {
	a[n] = a[n]
	  * (float)cos(M_PI * (n - m) * modfreq / fN);
      } else if (filterType == QFIL) {
	a[n] = a[n]
	  * (float)sin(M_PI * (n - m) * modfreq / fN);
      }
    }
    coeff = a;
    
    n_tap = order + 1;
    in_idx = 0;
    buf = (float *)(malloc(sizeof(float) * n_tap));
    this->filterGain();
    DOUT("order: %d\n", order);
  }

  ~FIRFilter() {
    free(coeff);
    free(buf);
  }
  void do_filter(float *x, int len);
  float do_filter(float x);
  void dump_coeff() {
    printf("number of coeffs: %d\n", n_tap);
    for (int i =0; i < n_tap; ++i) {
      printf("a[%d]: = %.7e\n", i, coeff[i]);
    }
  }
};

#endif
